/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.se.eip.patterns;

import java.net.URI;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.logging.Level;

import javax.jbi.JBIException;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;
import javax.xml.transform.dom.DOMSource;
import javax.xml.xpath.XPathExpressionException;

import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.util.UtilFactory;
import org.ow2.petals.se.eip.EIPConstants;
import org.ow2.petals.se.eip.ExchangeContext;
import org.ow2.petals.se.eip.async.CommonAsyncContext;
import org.w3c.dom.Document;

/**
 * Receives a bunch of messages and identifies messages that are correlated (XPATH expression). Once a complete set of messages has been received (XPATH expression on special message to decide that the set is complete), the aggregator collects information from each correlated message and publishes a single, aggregated
 * message to the output channel for further processing. It returns the result to this special message if it exists.
 * 
 * @author Adrien LOUIS - eBM WebSourcing
 * @author Christophe HAMERLING - eBM WebSourcing
 * @author Frederic Gardes
 * @since 2.0
 */
public class Aggregator extends AbstractPattern {

	public final static String AGGREGATOR_NAMESPACE = EIPConstants.NAMESPACE + "/aggregator";

	private Map<String, List<NormalizedMessage>> inMessagesMap;

	/*
	 * (non-Javadoc)
	 * @see org.ow2.petals.se.eip.patterns.AbstractPattern#process(org.ow2.petals .component.framework.util.Exchange, org.ow2.petals.se.eip.ExchangeContext)
	 */
	@Override
	public boolean process(final Exchange exchange, final ExchangeContext context) {
		try {
			final List<Consumes> consumes = context.getSUConsumes(exchange.getEndpoint());
			if (consumes != null && consumes.size() == 1) {
				final Document inMessageContent = UtilFactory.getSourceUtil().createDocument(exchange.getInMessage().getContent());
				// retrieve the document only once to avoid the Streaming issue
				exchange.getInMessage().setContent(new DOMSource(inMessageContent));
				final String correlationID = this.retrieveCorrelationId(inMessageContent, exchange, context);
				if (correlationID != null) {
					final boolean complete = this.isComplete(inMessageContent, context);
					if (this.inMessagesMap != null && !this.inMessagesMap.containsKey(correlationID) && !complete) {
						// TODO implement a timeout on stored exchanges
						this.inMessagesMap.put(correlationID, new CopyOnWriteArrayList<NormalizedMessage>());
					}
					if (complete) {
						if (this.logger != null && this.logger.isLoggable(Level.FINE)) {
							this.logger.fine(this.getName() + ": sequence is complete");
						}
						final Exchange response = this.sendAggregatedMessage(exchange, correlationID, context, consumes.get(0));
						if (response != null) {
							final NormalizedMessage receivedMessage = PatternHelper.processSentExchange(response, context);
							if (receivedMessage != null) {
								if (receivedMessage instanceof Fault && !exchange.isInOnlyPattern()) {
									exchange.setFault((Fault) receivedMessage);
								} else if (!exchange.isInOnlyPattern() && !exchange.isRobustInOnlyPattern()) {
									exchange.setOutMessage(receivedMessage);
								}
							}
						} else {
							exchange.setError(new JBIException("Timeout occurred, no reponse has been received from the target service"));
						}
					} else {
						if (this.logger != null && this.logger.isLoggable(Level.FINE)) {
							this.logger.fine(this.getName() + ": storing exchange '" + exchange.getExchangeId() + "' for correlation '" + correlationID + "'");
						}
						this.inMessagesMap.get(correlationID).add(exchange.getInMessage());
						if (this.logger != null && this.logger.isLoggable(Level.FINE)) {
							this.logger.fine(this.getName() + ": sequence size is : " + this.inMessagesMap.get(correlationID).size());
						}
						if (exchange.isInOutPattern() || exchange.isInOptionalOutPattern()) {
							final Document document = PatternHelper.createResultDocument(AGGREGATOR_NAMESPACE);
							document.getDocumentElement().setTextContent(this.getName() + ": the content is buffered by the pattern");
							exchange.getOutMessage().setContent(UtilFactory.getSourceUtil().createStreamSource(document));
						}
					}
				} else {
					exchange.setError(new Exception(this.getName() + ": no correlation id found in the message"));
				}
			} else {
				exchange.setError(new Exception(this.getName() + ": the Service Unit must define 1 valid Consumes section for the pattern "));
			}

		} catch (final JBIException e) {
			exchange.setError(new Exception(e));
		} catch (final XPathExpressionException e) {
			exchange.setError(new Exception(e));
		}
		return true;
	}

	/**
	 * Get the correlation ID from the message with an XPath expression or from the exchange property.
	 * 
	 * @return
	 * @throws JBIException
	 * @throws XPathExpressionException
	 * @throws XPathExpressionException
	 */
	private String retrieveCorrelationId(final Document inMessage, final Exchange exchange, final ExchangeContext context) throws XPathExpressionException {
		// retrieve the correlation from the message from an XPATH expression
		String result = null;
		final String expression = context.getExtensions().get(EIPConstants.AGGREGATOR_CORRELATION);
		if (expression != null) {
			result = PatternHelper.evaluateXPathExpressionAsString(expression, inMessage);
			if (result != null) {
				if (this.logger != null && this.logger.isLoggable(Level.FINE)) {
					this.logger.fine(this.getName() + ": evaluation of condition '" + expression + "', result '" + result + "'");
				}
				if (Boolean.FALSE.toString().equals(result) || result.length() == 0) {
					result = null;
				}
			}
		}
		return result;
	}

	/**
	 * @param exchange
	 * @return
	 * @throws JBIException
	 */
	private boolean isComplete(final Document document, final ExchangeContext context) throws JBIException {
		boolean result = false;
		final String condition = context.getExtensions().get(EIPConstants.TEST);
		try {
			result = PatternHelper.evaluateXPathCondition(condition, document.getDocumentElement());
			if (this.logger != null && this.logger.isLoggable(Level.FINE)) {
				this.logger.fine(this.getName() + ": evaluation of condition: \"" + condition + "\", result: \"" + Boolean.toString(result) + "\".");
			}
		} catch (final XPathExpressionException e) {
			throw new JBIException(this.getName() + ": error in XPathExpression '" + condition + "': " + e.getMessage());
		}
		return result;
	}

	/**
	 * @param exchange
	 * @param correlationId
	 * @param context
	 * @param consumes
	 * @throws JBIException
	 */
	private Exchange sendAggregatedMessage(final Exchange exchange, final String correlationId, final ExchangeContext context, final Consumes consumes) throws JBIException {
		final List<NormalizedMessage> inMessageList = this.inMessagesMap.get(correlationId);
		if (inMessageList == null) {
			throw new JBIException(this.getName() + ": no sequence found for the correlation " + correlationId);
		}
		// aggregate the messages into a single one...
		final Source content = this.buildContent(inMessageList, context);
		// send the aggregated message
		final Exchange response = this.sendExchange(exchange, correlationId, consumes, content, context);
		this.inMessagesMap.remove(correlationId);
		return response;
	}

	/**
	 * Build the aggregate Source from list of messages.
	 * 
	 * @throws JBIException
	 */
	private Source buildContent(final List<NormalizedMessage> inMessagesList, final ExchangeContext context) throws JBIException {
		Document aggregatedDocument = null;
		for (final NormalizedMessage inMessage : inMessagesList) {
			try {
				aggregatedDocument = PatternHelper.aggregateContent(inMessage, aggregatedDocument, EIPConstants.RESULT_ELEMENT, AGGREGATOR_NAMESPACE);
			} catch (final PEtALSCDKException e) {
				throw new JBIException(this.getName() + ": unable to aggregate the contents, " + e.getMessage());
			}
		}
		return UtilFactory.getSourceUtil().createDOMSource(aggregatedDocument);
	}

	/**
	 * Build the exchange to send to the target service and send it.
	 * 
	 * @param exchange
	 * @param correlationId
	 * @param consumes
	 * @param content
	 * @param context
	 * @return the result exchange
	 * @throws JBIException
	 */
	private Exchange sendExchange(final Exchange exchange, final String correlationId, final Consumes consumes, final Source content, final ExchangeContext context) throws JBIException {
		Exchange newExchange = context.createConsumeExchange(consumes);
		UtilFactory.getExchangeUtil().copyExchangeProperties(exchange, newExchange);
		final List<NormalizedMessage> inMessageList = this.inMessagesMap.get(correlationId);
		for (final NormalizedMessage inMessage : inMessageList) {
			PatternHelper.copy(inMessage, newExchange.getInMessage(), false, true, true, true);
		}
		final QName operation = consumes.getOperation();
		if (operation == null) {
			throw new JBIException(this.getName() + ": the consumes must have the operation of the targeted service");
		}
		newExchange.setOperation(operation);
		newExchange.setInMessageContent(content);
		this.logSend(consumes, newExchange);
		if (!context.sendSync(newExchange)) {
			newExchange = null;
		}
		return newExchange;
	}

	/*
	 * (non-Javadoc)
	 * @see org.ow2.petals.se.eip.patterns.AbstractPattern#processMEPValidation(java .net.URI)
	 */
	@Override
	public boolean validateMEP(final URI pattern) {
		return true;
	}

	/*
	 * (non-Javadoc)
	 * @see org.ow2.petals.se.eip.patterns.Pattern#init()
	 */
	public void init() {
		this.inMessagesMap = new ConcurrentHashMap<String, List<NormalizedMessage>>();
	}

	/*
	 * (non-Javadoc)
	 * @see org.ow2.petals.se.eip.patterns.Pattern#getName()
	 */
	public String getName() {
		return "Aggregator";
	}

	/*
	 * (non-Javadoc)
	 * @see org.ow2.petals.se.eip.patterns.AbstractPattern#processAsync(org.ow2.petals .component.framework.api.message.Exchange, org.ow2.petals.se.eip.ExchangeContext, org.ow2.petals.se.eip.async.CommonAsyncContext)
	 */
	@Override
	public boolean processAsync(final Exchange exchange, final ExchangeContext context, final CommonAsyncContext asyncContext) {
		exchange.setError(new Exception(this.getName() + ": the pattern doesn't process the asynchronous exchanges"));
		return true;
	}

	/*
	 * (non-Javadoc)
	 * @see org.ow2.petals.se.eip.patterns.AbstractPattern#getNameSpace()
	 */
	@Override
	public String getNameSpace() {
		return AGGREGATOR_NAMESPACE;
	}

}
