/**
 * PETALS - PETALS Services Platform. Copyright (c) 2008 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.se.eip.patterns;

import java.net.URI;
import java.util.List;

import javax.jbi.JBIException;
import javax.xml.transform.dom.DOMSource;
import javax.xml.xpath.XPathExpressionException;

import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.util.UtilFactory;
import org.ow2.petals.se.eip.EIPConstants;
import org.ow2.petals.se.eip.ExchangeContext;
import org.ow2.petals.se.eip.async.AbstractAsyncContext;
import org.ow2.petals.se.eip.async.CommonAsyncContext;
import org.w3c.dom.Document;

/**
 * A Router pattern on the content of the In message. No robustness on fault or
 * exception: this concept is linked to at least 2 target exchanges, 2 consumes
 * 
 * @author alouis
 * @author ofabre
 * @author Christophe HAMERLING - eBM WebSourcing
 * @author Frederic Gardes
 * @since 2.2
 */
public class Router extends AbstractMEPBridgeWithUnicConsumesPattern {

    public final static String ROUTER_NAMESPACE = EIPConstants.NAMESPACE + "/router";

    private RouterEngine routerEngine = null;

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#process(org.ow2.petals
     * .component.framework.util.Exchange,
     * org.ow2.petals.se.eip.ExchangeContext)
     */
    @Override
    public boolean process(final Exchange exchange, final ExchangeContext context) {
        try {
            final List<Consumes> consumesList = context.getSUConsumes(exchange.getEndpoint());
            final List<String> conditions = this.routerEngine.retrieveConditions(context
                    .getExtensions());
            // there must be numberOfConditions+1 consumes defined in the SU
            if (consumesList != null) {
                if (consumesList.size() == conditions.size() + 1) {
                    // reset the In content as a Document Source, to avoid the
                    // Streaming
                    // issue
                    final Document document = UtilFactory.getSourceUtil().createDocument(
                            exchange.getInMessage().getContent());
                    exchange.getInMessage().setContent(new DOMSource(document));
                    // retrieve identifier of the first consumes linked to a
                    // true
                    // condition, or the default one
                    final int targetedConsumesListIndentifier = this.routerEngine
                    .retrieveTargetedConsumesListIdentifier(conditions, document, context);
                    // forward the exchange to the correct service, depending on
                    // the
                    // test result
                    final Consumes consumedService = consumesList
                    .get(targetedConsumesListIndentifier);
                    final Exchange routedExchange = context.createConsumeExchange(consumedService);
                    PatternHelper.copy(exchange.getInMessage(), routedExchange.getInMessage());
                    UtilFactory.getExchangeUtil().copyExchangeProperties(exchange, routedExchange);
                    this.logSend(consumedService, routedExchange);
                    final boolean faultRobust = this.isFaultToException(context.getExtensions());
                    final CommonAsyncContext asyncContext = new CommonAsyncContext(exchange,
                            AbstractAsyncContext.DEFAULT_TIME_TO_LIVE, this.getName(), 1, false,
                            false, faultRobust);
                    context.sendAsync(routedExchange, asyncContext);
                } else {
                    exchange
                    .setError(new JBIException(
                            this.getName()
                            + ": the Service Unit must define (numberOfConditions+1) Consumes sections for the pattern "));
                }
            } else {
                exchange
                .setError(new JBIException(
                        this.getName()
                        + ": the Service Unit must define (numberOfConditions+2) Consumes sections for the pattern "));
            }

        } catch (final JBIException e) {
            exchange.setError(new Exception(e));
        } catch (final XPathExpressionException e) {
            exchange.setError(new Exception(e));
        }
        return false;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#getName()
     */
    public String getName() {
        return "Router";
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#validateMEP(java.net.URI)
     */
    @Override
    public boolean validateMEP(final URI mep) {
        return true;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#init()
     */
    public void init() {
        this.routerEngine = new RouterEngine();
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.AbstractPattern#getNameSpace()
     */
    @Override
    public String getNameSpace() {
        return ROUTER_NAMESPACE;
    }
}
