/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.se.eip.patterns;

import java.util.List;

import javax.jbi.JBIException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamSource;

import org.ow2.petals.commons.stream.InputStreamForker;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.util.UtilFactory;
import org.ow2.petals.se.eip.EIPConstants;
import org.ow2.petals.se.eip.ExchangeContext;
import org.ow2.petals.se.eip.async.AbstractAsyncContext;
import org.ow2.petals.se.eip.async.CommonAsyncContext;
import org.w3c.dom.Document;

/**
 * Aggregate all Consumes services defined in the service unit. <br>
 * Return the aggregated results.It accepts InOut only. If a Fault or Error
 * occurs on a call, stop the process and return the Fault or Error if no
 * robustness are used
 * 
 * @author alouis, chamerling - eBM WebSourcing
 * @author Frederic Gardes
 * 
 */
public class ScatterGather extends AbstractForkerPattern {

    private final static String SCATTER_GATHER_NAMESPACE = EIPConstants.NAMESPACE
    + "/scatter-gather";

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#process(org.ow2.petals
     * .component.framework.util.Exchange,
     * org.ow2.petals.se.eip.ExchangeContext)
     */
    @Override
    public boolean process(final Exchange exchange, final ExchangeContext context) {
        boolean reply = false;
        try {
            final List<Consumes> consumesList = context.getSUConsumes(exchange.getEndpoint());
            if (consumesList != null && consumesList.size() != 0) {
                this.sendAsyncExchanges(consumesList, exchange, context);
            } else {
                exchange
                .setError(new Exception(
                        this.getName()
                        + ": the Service Unit must define at least 1 Consumes sections for the pattern"));
                reply = true;
            }
        } catch (final JBIException e) {
            exchange.setError(new Exception(e));
            reply = true;
        }
        return reply;
    }

    /**
     * Send a copy of the exchange's IN message to all the services to consume.
     * Return the list of the Exchanges sent asynchronously. If an error occurs,
     * set a Fault on the original exchange, and return null
     * 
     * @param consumesList
     * @param exchange
     * @param context
     * @return the list of the exchanges sent async, or null if a problem
     *         occurred
     * @throws JBIException
     */
    private void sendAsyncExchanges(final List<Consumes> consumesList, final Exchange exchange,
            final ExchangeContext context) throws JBIException {
        Document document = null;
        InputStreamForker inMessageAsStream = null;
        final boolean faultRobust = this.isFaultRobust(context.getExtensions());
        final boolean exceptionRobust = this.isExceptionRobust(context.getExtensions());
        final CommonAsyncContext asyncContext = new CommonAsyncContext(exchange,
                AbstractAsyncContext.DEFAULT_TIME_TO_LIVE, this.getName(), consumesList.size(),
                faultRobust, exceptionRobust, false);
        final NormalizedMessage inMessage = exchange.getInMessage();
        // Handle the Streaming issue
        inMessageAsStream = PatternHelper.handleStreamSource(inMessage);
        if (inMessageAsStream == null) {
            if (inMessage.getContent() instanceof DOMSource) {
                // Handle the DOM issue
                document = UtilFactory.getSourceUtil().createDocument(inMessage.getContent());
            } else {
                throw new JBIException("The content must be a dom source or a stream source");
            }
        }
        // send all exchanges asynchronously
        for (final Consumes consumedService : consumesList) {
            final Exchange tmpExchange = context.createConsumeExchange(consumedService);
            PatternHelper.copy(inMessage, tmpExchange.getInMessage());
            UtilFactory.getExchangeUtil().copyExchangeProperties(exchange, tmpExchange);
            this.logSend(consumedService, tmpExchange);
            context.sendAsync(tmpExchange, asyncContext);
            if (inMessageAsStream != null) {
                // Handle the Streaming issue
                inMessage.setContent(new StreamSource(inMessageAsStream.getInputStreamTwo()));
                inMessageAsStream = PatternHelper.handleStreamSource(inMessage);
            } else {
                // Handle the DOM issue
                inMessage.setContent(UtilFactory.getSourceUtil().createDOMSource(document));
            }
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#init()
     */
    public void init() {
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#getName()
     */
    public String getName() {
        return "Scatter-Gather";
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.AbstractForkerPattern#getNameSpace()
     */
    @Override
    public String getNameSpace() {
        return SCATTER_GATHER_NAMESPACE;
    }
}
